#include<math.h>
#include<stdio.h>
#include<stdlib.h>
// second
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#define ZERO    0.0
#define ONE     1.0
#define EPS     1.0e-8

#define NUM     4096
#define BLK     16
#define DIS     256
//NUM = BLK*DIS

double  a[DIS][NUM][BLK];
double  b[DIS][BLK], ipvt[DIS][BLK], x[DIS][BLK];
int     rank, proc;

#ifdef _XCALABLEMP
#include<xmp.h>
#include<mpi.h>
#pragma xmp nodes p(*)
#pragma xmp template t(0:256-1) //DIS-1
#pragma xmp distribute t(cyclic) onto p
#pragma xmp align a[i][*][*] with t(i)
#pragma xmp align b[i][*] with t(i)
#pragma xmp align x[i][*] with t(i)
#pragma xmp align ipvt[i][*] with t(i)
#endif

double
second(void)
{
  double t;
#ifdef USE_GETRUSAGE
  struct rusage ru;
  void getrusage();

  getrusage(RUSAGE_SELF,&ru);
  t = (double) (ru.ru_utime.tv_sec+ru.ru_stime.tv_sec) +
    ((double) (ru.ru_utime.tv_usec+ru.ru_stime.tv_usec))/1.0e6 ;
  return t ;
#else
  struct timeval tv;
  gettimeofday(&tv,NULL);
  t = (double) (tv.tv_sec) + ((double) (tv.tv_usec))/1.0e6;
  return t ;
#endif
}
// second

//係数行列と右辺ベクトルの初期化. norma = ||A||
void 
matgen(double a[DIS][NUM][BLK], 
       double b[DIS][BLK], double *norma)
{
  int           init, i, j, k;
  double        norma_temp;
#ifdef _XCALABLEMP
  rank = xmp_get_rank();
#else
  rank = 0;
#endif
  init   = 1325;
  *norma = 0.0;

  srand48(rank);

#pragma xmp loop on t(k)
  for(k=0; k<DIS; k++){
    for(i=0; i<NUM; i++){
      for(j=0; j<BLK; j++){
        init = 3125*init%65536;
        a[k][i][j] = drand48(); //(init-32768.0)/16384.0;
        norma_temp = (a[k][i][j] > norma_temp)?
	  a[k][i][j]:norma_temp;
      }
    }
  }
#pragma xmp reduction(MAX:norma_temp) on p
  *norma = norma_temp;

#pragma xmp loop on t(k)
  for(k=0; k<DIS; k++){
    for(j=0; j<BLK; j++){
      b[k][j] = 0.0;
    }
  }

#pragma xmp loop on t(k)
  for(k=0; k<DIS; k++){
    for(i=0; i<NUM; i++){
      for(j=0; j<BLK; j++){
        b[k][j] += a[k][i][j];
      }
    }
  }

}
/*--- 係数行列と右辺ベクトルの初期化. norma = ||A|| ---*/

/*--- AをLU分解する. ---*/
void
dgefa(double a[DIS][NUM][BLK], int ipvt[DIS][BLK])
{
  int           i,j,k;
  int           l,r, s;
  int           run_now, btop;
  double        temp, sum;
  double        pvt_v[BLK][NUM];

#ifdef _XCALABLEMP
  proc = xmp_get_size();
#else
  proc = 1;
#endif

  for(k=0; k<DIS; k++){

    run_now = 1 + k%proc; //cyclic
#pragma xmp task on p(run_now)
    {//現ブロックでのLU分解
      for(i=0; i<BLK; i++){
        //ブロック内で最大値探索
        temp = 0.0;
        l = i;
        r = k*BLK + i;
        for(j=i; j<BLK;j++){
          if(temp < fabs(a[k][r][j]) ){
            l = j;
            temp = fabs(a[k][r][j]);
          }
        }
        ipvt[k][i] = l;
        //ピボット入れ替え
        for(j=r; j<NUM; j++) pvt_v[i][j] = a[k][j][l];
        if(l != i){
          for(j=r; j<NUM; j++) a[k][j][l] = a[k][j][i];
          for(j=r; j<NUM; j++) a[k][j][i] = pvt_v[i][j];
        }
        temp    = -ONE/pvt_v[i][r];
        //i行目の更新
        for(j=i+1;j<BLK;j++) a[k][r][j] *= temp;

        //i+1 - BLK 間の列の更新
        for(j=i+1; j<BLK; j++){
          for(s=r+1; s<NUM; s++){
            a[k][s][j] += pvt_v[i][s]*a[k][r][j];
          }
        }
      }
    }//現ブロック

/* #pragma xmp bcast (pvt_v) from p(run_now) */
    MPI_Bcast(pvt_v, BLK*NUM, MPI_DOUBLE, run_now-1, MPI_COMM_WORLD);
    //各ブロックの更新

#pragma xmp loop on t(l)
    for(l=k+1; l<DIS; l++){
      for(s=k*BLK; s<(k+1)*BLK; s++){
        btop = s%BLK;
        //U
        temp    = -ONE/pvt_v[btop][s];
        for(j=0; j<BLK; j++){
          a[l][s][j] *= temp;
        }
        //L
        for(j=0; j<BLK; j++){
          temp = a[l][s][j];
          for(i=s+1;i<NUM;i++){
            a[l][i][j] += pvt_v[btop][i]*temp;
          }
        }
      }
    }
    //各ブロックの更新
  }
}
/*--- AをLU分解する. ---*/

/*--- LU分解したAを用いてAx=bを解く. bに解が格納される. ---*/
void
dgesl(double a[DIS][NUM][BLK], int ipvt[DIS][BLK],
      double b[DIS][BLK])
{
  int           i,j,k;
  int           l, s, btop;
  int           run_now;
  double        temp, sum;

#ifdef _XCALABLEMP
  proc = xmp_get_size();
#else
  proc = 1;
#endif

  for(k=0; k<DIS; k++){
    run_now = 1 + k%proc; //cyclic
    //kブロックでのbベクトルのピボット入れ替え
    for(i=0; i<BLK; i++){
#pragma xmp task on p(run_now)
      {
        l = ipvt[k][i];
        temp = b[k][l];
        if(l != i){
          b[k][l] = b[k][i];
          b[k][i] = temp;
        }
        for(j=i+1; j<BLK; j++)
          b[k][j] += temp*a[k][k*BLK+i][j];
      }

      //k+1ブロック以降の計算
#pragma xmp bcast (temp) from p(run_now)
#pragma xmp loop on t(l)
      for(l=k+1; l<DIS; l++){
        for(s=0; s<BLK; s++)
          b[l][s] += temp*a[l][k*BLK+i][s];
      }
    }
  }

  for(k=DIS-1; k>=0; k--){
    run_now = 1 + k%proc; //cyclic
    for(i=BLK-1; i>=0; i--){
#pragma xmp task on p(run_now)
      {
        b[k][i] = b[k][i]/a[k][k*BLK+i][i];
        temp = -b[k][i];
        for(j=0; j<i; j++)
          b[k][j] += temp*a[k][k*BLK+i][j];
      }
#pragma xmp bcast (temp) from p(run_now)
#pragma xmp loop on t(l)
      for(l=0; l<k; l++){
        for(s=0; s<BLK; s++)
          b[l][s] += temp*a[l][k*BLK+i][s];
      }
    }
  }
}
/*--- LU分解したAを用いてAx=bを解く. bに解が格納される. ---*/

double
epslon(double x)
{
  double a, b, c, eps;
  a = 4.0e0/3.0e0;
  eps = ZERO;

  while (eps == ZERO) {
    b = a - ONE;
    c = b + b + b;
    eps = fabs(c-ONE);
  }

  return(eps*fabs(x));
}

/*--- 解の評価 ---*/
void 
check_sol(double a[DIS][NUM][BLK],
          double b[DIS][BLK], double x[DIS][BLK])
{
  int           i, j, k;
  double        norma, normx, residn, resid,eps, temp;
  double        y[DIS][BLK];

#pragma xmp loop on t(i)
  for(i=0; i<DIS; i++){
    for(j=0; j<BLK; j++){
      x[i][j] = b[i][j];
    }
  }
  rank = xmp_get_rank();
  MPI_Allgather(x,NUM/proc,MPI_DOUBLE,
                y,NUM/proc,MPI_DOUBLE, MPI_COMM_WORLD);

  matgen((void*)a, b, &norma);

#pragma xmp loop on t(i)
  for(i=0; i<DIS; i++){
    for(j=0; j<BLK; j++){
      b[i][j] = -b[i][j];
    }
  }
#pragma xmp loop on t(k)
  for(k=0; k<DIS; k++){
    for(j=0; j<BLK; j++){
      temp = 0.0;
      for(i=0; i<NUM; i++)
        temp += a[k][i][j]*y[k][i%BLK];
      b[k][j] += temp;
    }
  }

  resid = 0.0;
  normx = 0.0;

#pragma xmp loop on t(k)
  for(k=0; k<DIS; k++){
    for(j=0; j<BLK; j++){
      resid = (resid > fabs((double)b[k][j]))?
        resid : fabs((double)b[k][j]);
      normx = (normx > fabs((double)x[k][j]))? 
        normx : fabs((double)x[k][j]);
    }
  }
#pragma xmp reduction(MIN:resid, normx) on p

  eps = epslon((double)ONE);
  residn = resid/( NUM*norma*normx*eps );

#pragma xmp task on p(1)
  {
    printf("   norm. resid      resid           machep");
    printf("         x[0]-1        x[n-1]-1\n");
    printf("%8.1f      %16.8e%16.8e%16.8e%16.8e\n",
           (double)residn, (double)resid, (double)eps, 
           (double)y[0][0]-1, (double)y[DIS-1][BLK-1]-1);
  }
}
/*--- 解の評価 ---*/

int 
main(int argc, char* argv[])
{
  int           i,j,k;
  double        ops,norma, t0,t1,t, dn;

#ifdef _XCALABLEMP
  rank = xmp_get_rank();
  proc = xmp_get_size();
#else
  rank = 0;
  proc = 1;
#endif

  dn = (double)NUM;
  ops = (2.0e0*(dn*dn*dn))/3.0 + 2.0*(dn*dn);

  if(rank == 0){
    printf("Linpack ...\nN:%d proc:%d\n", (int)dn, proc);
  }

  matgen(a, b, &norma);

  t0 = second();
  dgefa(a,ipvt);
  dgesl(a,ipvt, b);
  t1 = second();

  t = t1 - t0;
#pragma xmp reduction(MAX:t) on p
  if(rank == 0)
    printf("time=%g, %g GFlops\n", t,ops/(t*1.0e9));
  if(rank == 0)
    printf("check...\n");
  check_sol(a, b, x);

  if(rank == 0)
    printf("end ...\n");

  return 0;
}
